import os
import logging
import time
import h5py

import gonzales.lib.brute_force as bf
import gonzales.lib.octree as oct
import gonzales.lib.physics as phy
from gonzales.simulator.utils import ProgressBar

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)


# ------------------------------------------------------
# Result reader
# ------------------------------------------------------

class ResultReader:
    """
    Reader for hdf5 results. Should be closed after using.
    """
    def __init__(self, hdf5_filepath):
        """

        :param hdf5_filepath: path to hdf5 file
        """
        if not h5py.is_hdf5(hdf5_filepath):
            raise ValueError('Not an hdf5 file: {}'.format(hdf5_filepath))
        self._fobj = h5py.File(hdf5_filepath, 'r')

    def get_info(self):
        """
        Returns simulation info, which is dict-like iterable.
        """
        return self._fobj['info']

    def get_result(self, result_name, step_num):
        """
        Get a specific result for a given step.

        :param result_name: str, name of result, must be supported by simulation
        :param step_num: step number
        """
        return self._fobj['results'][result_name][step_num]

    def get_result_frequency(self, result_name):
        """
        Get result frequency.

        :param result_name: str, name of result, must be supported by simulation
        
        Returns result frequency
        """        
        return self._fobj['results'][result_name].attrs['frequency']

    def get_result_num_steps(self, result_name):
        """

        :param result_name: str, name of result, must be supported by simulation
        
        Returns number of steps for which result is calculated
        """
        return self._fobj['results'][result_name].shape[0]

    def get_result_names(self):
        return list(self._fobj['results'].keys())

    def close(self):
        """
        Closes the hdf5 file. This should be called when the reader is not needed anymore.
        """
        if bool(self._fobj):
            self._fobj.close()


# ------------------------------------------------------
# Simulation classes
# ------------------------------------------------------

class ResultDesc:
    def __init__(self, shape, writer_func, frequency):
        self.shape = shape
        self.writer_func = writer_func
        self.frequency = frequency


class SimulationBase:
    def __init__(self, space, output_filepath, G, eps):
        """
        Base class for all simulation types.

        :param space: instance of Space
        :param output_filepath: output filepath
        :param G: gravitational constant
        :param eps: gravitational softening
        """
        self.space = space        
        self.output_filepath = output_filepath
        self._ensure_output_dir()
        self.G = G
        self.eps = eps
        self._pb = ProgressBar(1, 40)
        self._kernel = None
        self._result_descs = {'position': ResultDesc(self.space.r.shape, self._write_position, 1),
                              'velocity': ResultDesc(self.space.v.shape, self._write_velocity, 1),
                              'energy': ResultDesc((1,), self._write_total_energy, 0),
                              'angular_momentum': ResultDesc((3,), self._write_angular_momentum, 0)}

    def _ensure_output_dir(self):
        head, tail = os.path.split(self.output_filepath)
        if not head:
            return
        if not os.path.isdir(head):
            try:
                os.mkdir(head)
            except:
                raise OSError('Error creating directory: {}'.format(head))

    def add_result(self, res_name, res_frequency=1):
        """
        Adds a result which shall be written to the hdf5 file during the simulation. The shape
        of the result needs to be provided, as does the frequency of the output. The default res_frequency
        1 means the result will be written to the file at each simulation step.

        :param res_name: string, name of result
        :param res_frequency: int, result will be written every res_frequency steps. If 0, result will not
            be written at all.
        :return:
        """
        assert res_name in self._result_descs, 'Invalid result'
        self._result_descs[res_name].frequency = res_frequency

    def _write_position(self, hdf5_fobj, step_num):
        hdf5_fobj['results/position'][step_num, :] = self.space.r

    def _write_velocity(self, hdf5_fobj, step_num):
        hdf5_fobj['results/velocity'][step_num, :] = self.space.v

    def _write_total_energy(self, hdf5_fobj, step_num):
        hdf5_fobj['results/energy'][step_num] = phy.calc_te(self.space.r, self.space.v, self.space.m, self.G, self.eps)

    def _write_angular_momentum(self, hdf5_fobj, step_num):
        hdf5_fobj['results/angular_momentum'][step_num] = phy.calc_ang_mom(
            self.space.r, self.space.v, self.space.m)

    def _write_results(self, hdf5_fobj, step_num):
        for res_name, res_data in self._result_descs.items():
            res_freq, res_shape = res_data.frequency, res_data.shape
            if res_freq == 0:
                continue
            if step_num % res_freq == 0:
                res_data.writer_func(hdf5_fobj, int(step_num / res_freq))

    @staticmethod
    def set_metadata(hdf5_obj, **kwargs):
        """
        Sets metadata attributes for hdf5 object (can be a group or a dataset)
        """
        for k, v in kwargs.items():
            hdf5_obj[k] = v

    def create_datasets(self, hdf5_fobj, n_steps, step_size):
        """
        Creates hdf5 datasets.
        """
        info_grp = hdf5_fobj.create_group('info')
        self.set_metadata(info_grp, number_of_steps=n_steps, time_step_size=step_size, G=self.G, epsilon=self.eps,
                          start_time=time.time(), number_of_particles=len(self.space), simulation_type=self.type)
        results_grp = hdf5_fobj.create_group('results')
        for res_name, res_data in self._result_descs.items():
            res_freq, res_shape = res_data.frequency, res_data.shape
            if res_freq == 0:
                continue
            n_rows = int(n_steps / res_freq)
            results_grp.create_dataset(res_name, (n_rows + 1, *res_shape))
            results_grp[res_name].attrs['frequency'] = res_freq

    def set_kernel(self, kernel_func):
        """
        Sets the kernel function which will be used to calculate accelerations for each simulation
        step.

        :param kernel_func: function used to calculate accelerations
        """
        self._kernel = kernel_func

    def calc_accs(self):
        raise NotImplementedError

    def run(self, n_steps, step_size):
        """
        Runs the simulation.

        :param n_steps: number of steps
        :param step_size: step size
        """
        # reset Progress bar
        self._pb.reset(n_steps)
        with h5py.File(self.output_filepath, 'w') as res_f:
            # set initial hdf5 data
            logging.info('Simulation - type={}, N_particles={}, N_steps={}'.format(self.type, len(self.space),
                                                                                   n_steps))
            logging.info('Creating datasets')
            self.create_datasets(res_f, n_steps, step_size)
            logging.info('Writing initial data')
            self._write_results(res_f, 0)
            # calculate initial accelerations
            logging.info('Calculating initial accelerations')
            accs = self.calc_accs()
            logging.info('Start simulation')
            # integration
            for i in range(1, n_steps + 1):
                phy.advance_r(self.space.r, self.space.v, accs, step_size)
                new_accs = self.calc_accs()
                phy.advance_v(self.space.v, accs, new_accs, step_size)
                accs = new_accs
                # update hdf5 data
                self._write_results(res_f, i)
                self._pb.update()
            res_f['info']['end_time'] = time.time()
            res_f['info']['total_time'] = res_f['info']['end_time'][()] - res_f['info']['start_time'][()]
        logging.info('End simulation')


class PPSimulation(SimulationBase):
    """
    Simulation class for brute-force simulations.
    """
    def __init__(self, *args, **kwargs):
        """
        :param space: instance of Space
        :param output_filepath: output filepath
        :param G: gravitational constant
        :param eps: gravitational softening
        """
        super().__init__(*args, **kwargs)
        self.type = 'Brute force'
        self.set_kernel(bf.calculate_accs_pp)

    def __repr__(self):
        return '<{}[{}] N={}, type={}>'.format(type(self).__name__, id(self), len(self.space), self.type)

    def calc_accs(self):
        """
        Calculates the acceleration vector, using the kernel which was set.
        """
        return self._kernel(self.space.r, self.space.m, self.G, self.eps)


class BHSimulation(SimulationBase):
    """
    Simulation class for Barnes-Hut simulations.
    """
    def __init__(self, space, output_filepath, G, eps, root_width, root_center, theta):
        """
        :param space: instance of Space
        :param output_filepath: output filepath
        :param G: gravitational constant
        :param eps: gravitational softening
        :param root_width: octree root node width
        :param root_center: octree root node center
        :param theta: threshold parameter
        """
        super().__init__(space, output_filepath, G, eps)
        self.type = 'Barnes-Hut'
        self.root_width = root_width
        self.root_center = root_center
        self.theta = theta
        self.set_kernel(oct.calc_accs_octree)

    def __repr__(self):
        return '<{}[{}] N={}, type={}>'.format(type(self).__name__, id(self), len(self.space), self.type)

    def calc_accs(self):
        """
        Calculates the acceleration vector, using the kernel which was set.
        """
        return self._kernel(self.root_width, *self.root_center, self.space.r, self.space.m, self.G, self.eps,
                            self.theta)



import h5py
import numpy as np

from gonzales.simulator.utils import to_cartesian


class Space:
    """
    Class used to represent a 3D space which contains particles.
    """
    def __init__(self, r=None, v=None, m=None):
        """
        Initialize the space.

        :param r: N x 3 numpy array or None, position vectors of the particles
        :param v: N x 3 numpy array or None, velocity vectors of the particles
        :param m: N x 1 numpy array or None, mass vector
        """
        self.r = np.empty(shape=(0, 3)) if r is None else r
        self.v = np.empty(shape=(0, 3)) if v is None else v
        self.m = np.array([]) if m is None else m

    def to_hdf5(self, filepath):
        """
        Saves configuration to hdf5 file. Datasets are called "r", "v" and "m".
        """
        with h5py.File(filepath, 'w') as f:
            f.create_dataset('r', self.r.shape)
            f.create_dataset('v', self.v.shape)
            f.create_dataset('m', self.m.shape)
            f['r'][:] = self.r
            f['v'][:] = self.v
            f['m'][:] = self.m

    def from_hdf5(self, filepath):
        """
        Reads data from hdf5 file.
        """
        with h5py.File(filepath, 'r') as f:
            self.add_particles(f['r'][:], f['v'][:], f['m'][:])

    def add_particle(self, r, v, m):
        """
        Add a single particle with the specified position, velocity and mass.

        :param r: numpy array, position vector of the particle
        :param v: numpy array, velocity vector of the particle
        :param m: float, particle mass
        """
        self.r = np.vstack((self.r, r))
        self.v = np.vstack((self.v, v))
        self.m = np.append(self.m, m)

    def add_particles(self, r, v, m):
        """
        Adds N particles, specified by position and velocity matrices, and by mass vector.

        :param r: N x 3 numpy array, position vectors of the particles
        :param v: N x 3 numpy array, velocity vectors of the particles
        :param m: N x 1 numpy array, mass vector
        """
        self.r = np.vstack((self.r, r))
        self.v = np.vstack((self.v, v))
        self.m = np.append(self.m, m)

    def add_cuboid(self, n, center, l_x, l_y, l_z, v_func, m_func):
        """
        Generates uniform random particle distribution within a cuboid volume
        with given center and side lengths.

        :param n: number of particles to generate
        :param center: center of cuboid volume
        :param l_x: length of cuboid in the x-direction
        :param l_y: length of cuboid in the y-direction
        :param l_z: length of cuboid in the z-direction
        :param v_func: function of position vector
        :param m_func: function of position vector
        """
        r = np.random.uniform(
            np.array([-l_x, -l_y, -l_z]) * 0.5, np.array([l_x, l_y, l_z]) * 0.5,
            (n, 3))
        v = np.apply_along_axis(v_func, 1, r)
        self.r = np.vstack((self.r, r + center))
        self.v = np.vstack((self.v, v))
        self.m = np.append(self.m, np.apply_along_axis(m_func, 1, r))

    def add_sphere(self, n, center, radius, v_func, m_func):
        """
        Generates uniform random particle distribution within a spherical volume
        with a given center and radius.

        :param n: number of particles to generate
        :param center: center of spherical volume
        :param radius: radius of spherical volume
        :param v_func: function of r, theta, phi
        :param m_func: function of r, theta, phi
        """
        l = np.random.uniform(0, 1, n)
        u = np.random.uniform(-1, 1, n)
        r = radius * l ** (1/3)
        theta = np.arccos(-u)
        phi = np.random.uniform(0, 2 * np.pi, (n,))
        r_spherical = np.column_stack((r, theta, phi))
        v_spherical = np.apply_along_axis(v_func, 1, r_spherical)
        r_cartesian = to_cartesian(r_spherical, 'spherical')
        v_cartesian = to_cartesian(v_spherical, 'spherical')
        self.r = np.vstack((self.r, r_cartesian + center))
        self.v = np.vstack((self.v, v_cartesian))
        self.m = np.append(
            self.m, np.apply_along_axis(m_func, 1, r_spherical))

    def add_cylinder(self, n, center, radius, l_z, v_func, m_func):
        """
        Generates uniform random particle distribution within a cylindrical
        volume with a given center, radius and height.

        :param n: number of particles to generate
        :param center: center of cylindrical volume
        :param radius: radius of cylindrical volume
        :param l_z: height of cylindrical volume
        :param v_func: function of position vector in cylindrical coordinates
        :param m_func: function of position vector in cylindrical coordinates
        """
        r = np.random.uniform(0, radius, n)
        theta = np.random.uniform(0, 2 * np.pi, n)
        z = np.random.uniform(-0.5 * l_z, 0.5 * l_z, n)
        x = np.sqrt(r) * np.cos(theta)
        y = np.sqrt(r) * np.sin(theta)
        r_cylindrical = np.column_stack((r, theta, z))
        v_cylindrical = np.apply_along_axis(v_func, 1, r_cylindrical)
        r_cartesian = np.column_stack((x, y, z))
        v_cartesian = to_cartesian(v_cylindrical, 'cylindrical')
        self.r = np.vstack((self.r, r_cartesian + center))
        self.v = np.vstack((self.v, v_cartesian))
        self.m = np.append(
            self.m, np.apply_along_axis(m_func, 1, r_cylindrical))

    def add_plummer(self, n, center):
        m = np.ones(n) * 1. / n
        self.m = np.append(self.m, m)
        r = 1. / np.sqrt(np.random.rand(n)**(-2. / 3.) - 1)
        phi = np.random.uniform(0, 2 * np.pi, n)
        theta = np.arccos(np.random.uniform(-1, 1, n))
        R = np.ones((n, 3))
        R[:, 0] = np.multiply(np.multiply(r, np.sin(theta)), np.cos(phi))
        R[:, 1] = np.multiply(np.multiply(r, np.sin(theta)), np.sin(phi))
        R[:, 2] = np.multiply(r, np.cos(theta))
        self.r = np.vstack((self.r, R + center))
        x = np.zeros(n)
        y = np.ones(n) * 0.1
        while all(y > x**2 * (1. - x**2)**3.5):
            x = np.random.uniform(0, 1, n)
            y = np.random.uniform(0, 0.1, n)
        phi = np.random.uniform(0, 2 * np.pi, n)
        theta = np.arccos(np.random.uniform(-1, 1, n))
        vel = np.multiply(x, np.sqrt(2) * (1 + np.multiply(r, r)) ** -0.25)
        V = np.ones((n, 3))
        V[:, 0] = np.multiply(np.multiply(vel, np.sin(theta)), np.cos(phi))
        V[:, 1] = np.multiply(np.multiply(vel, np.sin(theta)), np.sin(phi))
        V[:, 2] = np.multiply(vel, np.cos(theta))
        self.v = np.vstack((self.v, V))

    def clear_particles(self):
        """
        Removes all particles from the space.
        """
        self.r = np.empty(shape=(0, 3))
        self.v = np.empty(shape=(0, 3))
        self.m = np.array([])

    def __bool__(self):
        return np.any(self.m)

    def __len__(self):
        return len(self.m)



import platform
import psutil
import timeit
import logging
import numpy as np
import functools
import json

from gonzales.simulator.space import Space
import gonzales.lib.brute_force as bf
import gonzales.lib.octree as oct


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)


def vel_func(pos_vec):
    return np.array((0., 0., 0.))


def mass_func(pos_vec):
    return 1.0


class BenchmarkConfigBase:
    """
    Base class for benchmark configuration.
    """
    def __init__(self, particle_nums, num_exec=5):
        """
        Initialize configuration.

        :param particle_nums: tuple of integers representing particle numbers to be generated
        :param num_exec: number of times the function is timed
        """
        self.particle_nums = particle_nums
        self.num_exec = num_exec


class PPBenchmarkConfig(BenchmarkConfigBase):
    """
    Benchmark configuration for brute force acceleration calculation
    """
    def do_timing(self, space):
        """
        Executes the calculation.

        :param space: instance of gonzales.simulator.space
        """
        t = timeit.Timer(functools.partial(bf.calculate_accs_pp, space.r, space.m, 1.0, 0.))
        res = t.timeit(self.num_exec) / self.num_exec
        return res


class BHBenchmarkConfig(BenchmarkConfigBase):
    """
    Benchmark configuration for Barnes-hut acceleration calculation
    """
    def __init__(self, particle_nums, theta=0.75, num_exec=5):
        """
        Initialize configuration.

        :param particle_nums: tuple of integers representing particle numbers to be generated
        :param theta: Barnes-Hut parameter
        :param num_exec: number of times the function is timed
        """
        super().__init__(particle_nums, num_exec)
        self.theta = theta

    def do_timing(self, space):
        """
        Executes the calculation.

        :param space: instance of gonzales.simulator.space
        """
        t = timeit.Timer(functools.partial(oct.calc_accs_octree, 1., 0., 0., 0., space.r, space.m, 1.0, 0., self.theta))
        res = t.timeit(self.num_exec) / self.num_exec
        return res


class BenchmarkSuite:
    """
    Benchmark suite.
    """
    def __init__(self, output_filename=None):
        """
        Initialize the suite.

        :param output_filename: filename of output json file
        """
        self.output_filename = output_filename if output_filename is not None else 'benchmark_results.json'
        self.benchmarks = {}
        self._results = {}

    @staticmethod
    def _get_system_info():
        return {'platform': platform.system(),
                'platform_release': platform.release(),
                'platform_version': platform.version(),
                'processor': platform.processor(),
                'cores': psutil.cpu_count(),
                'memory': round(psutil.virtual_memory().total / (1024.0**3))}

    def add_benchmark(self, bench_name, bench_config):
        """
        Adds a benchmark instance to the suite.

        :param bench_name: str, name of benchmark
        :param bench_config: instance of BenchmarkConfigBase
        """
        self.benchmarks[bench_name] = bench_config

    @staticmethod
    def _generate_space(num):
        space = Space()
        space.add_cuboid(num, np.array((0., 0., 0.)), 1., 1., 1., vel_func, mass_func)
        return space

    def run(self):
        """
        Runs the benchmark suite, logs information and dumps results to json.
        """
        logging.info('Gathering system information')
        self._results['sys_info'] = self._get_system_info()
        self._results['benchmarks'] = dict()
        logging.info('Start benchmark suite')
        for bench_name, bench_config in self.benchmarks.items():
            self._results['benchmarks'][bench_name] = dict()
            for num in bench_config.particle_nums:
                space = self._generate_space(num)
                logging.info('Running benchmark: {}, N={}'.format(bench_name, num))
                self._results['benchmarks'][bench_name][num] = bench_config.do_timing(space)
        logging.info('Dumping results to json')
        self._dump_to_json()
        logging.info('End benchmark suite')

    def _dump_to_json(self):
        with open(self.output_filename, 'w') as fp:
            json.dump(self._results, fp)


def run_default_benchmark(output_filename='benchmark_results.json'):
    """
    Runs the default benchmark suite.

    :param output_filename: str, name of output json file
    """
    bf_config = PPBenchmarkConfig((1000, 10000, 20000, 30000))
    bh_config = BHBenchmarkConfig((10000, 100000, 500000, 1000000))
    suite = BenchmarkSuite(output_filename=output_filename)
    suite.add_benchmark('brute_force', bf_config)
    suite.add_benchmark('barnes_hut', bh_config)
    suite.run()



import numpy as np

import gonzales.lib.physics as phy


def test_com():
    # test center of mass calculation
    # 1. case
    r = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    m = np.array([1., 2., 3.])
    com = phy.calc_com(r, m)
    np.testing.assert_almost_equal(com, np.array([5/6., 0.5, 0.5]))
    # 2. case - one of the masses dominates
    r = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    m = np.array([1., 2., 1.0e15])
    com = phy.calc_com(r, m)
    np.testing.assert_almost_equal(com, np.array([1., 1., 1.]))


def test_ke():
    # test kinetic energy calculation
    # 1. trivial case
    v = np.array([[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])
    m = np.array([1., 2., 3.])
    ke = phy.calc_ke(v, m)
    np.testing.assert_equal(ke, 0.)
    # 2. other cases
    v = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    m = np.array([1., 2., 3.])
    ke = phy.calc_ke(v, m)
    np.testing.assert_equal(ke, 11 / 2.)


def test_pe():
    # test potential energy calculation
    # 1. case
    r = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    m = np.array([1., 2., 3.])
    pe = phy.calc_pe(r, m, 1.0, 0.)
    np.testing.assert_equal(pe, - (2. + 3 / np.sqrt(3) + 6 / np.sqrt(2)))
    # 2. case - particles at huge distances
    r = np.array([[0., 0., 0.], [1.0e15, 0., 0.], [1., 1.0e15, 1.]])
    m = np.array([1., 2., 3.])
    pe = phy.calc_pe(r, m, 1.0, 0.)
    np.testing.assert_almost_equal(pe, 0.)


def test_te():
    # test total energy calculation
    r = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    v = np.array([[-1., 1., 0.], [1., -1., 0.], [1., 1., 1.]])
    m = np.array([1., 2., 3.])
    pe = phy.calc_pe(r, m, 1.0, 0.)
    ke = phy.calc_ke(v, m)
    te = phy.calc_te(r, v, m, 1.0, 0.)
    np.testing.assert_equal(pe + ke, te)


def test_ang_mom():
    # test angular momentum
    r = np.array([[0., 0., 0.], [1., 0., 0.], [1., 1., 1.]])
    v = np.array([[-1., 1., 0.], [1., -1., 0.], [1., -2., 1.]])
    m = np.array([1., 2., 3.])
    am = phy.calc_ang_mom(r, v, m)
    np.testing.assert_equal(am, np.array([9., 0., -11.]))



cimport cython
cimport numpy as np
import numpy as np
from libc.math cimport sqrt

ctypedef np.float64_t DTYPE_t


cdef extern from "octnode.h":
	double* calc_accs_wrap(int n, double* points, double* masses, double G, double eps, double theta, double root_width, double root_x, double root_y, double root_z)


cdef _calc_accs(int n, DTYPE_t [:, :] r, DTYPE_t [:] m, double G, double eps, double theta, double w, double r_x, double r_y, double r_z):
	cdef double* accs
	accs = calc_accs_wrap(n, &r[0, 0], &m[0], G, eps, theta, w, r_x, r_y, r_z)
	accs_py = np.asarray(<DTYPE_t[:n, :3]> accs)
	return accs_py


def calc_accs_octree(w, r_x, r_y, r_z, r, m, G, eps, theta):
	"""
    Parallel calculation of accelerations using the Barnes-Hut algorithm. Constructs the
	octree and calculates the accelerations.

	:param w: double, root node width
	:param r_x: double, x-coordinate of root node center
	:param r_y: double, y-coordinate of root node center
	:param r_z: double, z-coordinate of root node center
	:param r: N x 3 numpy array, position vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:param G: gravitational constant
	:param eps: gravitational softening
	:param theta: threshold parameter
	:return: N x 3 numpy array, acceleration vectors of particles
	"""
	return _calc_accs(r.shape[0], r, m, G, eps, theta, w, r_x, r_y, r_z)



import numpy as np

from gonzales.simulator.space import Space
from gonzales.simulator.utils import calculate_relative_error
import gonzales.lib.brute_force as bf
import gonzales.lib.octree as oct


def vel_func(pos_vec):
    return np.array((0, 0, 0))


def mass_func(pos_vec):
    return 1.0


def test_brute_force_basic():
    # test correctness of Cython kernel
    G = 1.0
    eps = 0
    space = Space()
    # 1. zero acceleration cases
    # 1.1. single particle
    space.add_particle(np.array((0., 0., 0.)), np.array((0., 0., 0.)), 1.0)
    accs = bf.calculate_accs_pp(space.r, space.m, G, eps)
    np.testing.assert_equal(accs, np.array([[0., 0., 0.]]))
    # 1.2. two particles at big distance
    space.add_particle(np.array((1.0e10, 1.0e15, 1.0e15)), np.array((0., 0., 0.)), 1.0)
    accs = bf.calculate_accs_pp(space.r, space.m, G, eps)
    np.testing.assert_almost_equal(accs, np.array([[0., 0., 0.], [0., 0., 0.]]))
    # 1.3. set G = 0
    accs = bf.calculate_accs_pp(space.r, space.m, 0., eps)
    np.testing.assert_equal(accs, np.array([[0., 0., 0.], [0., 0., 0.]]))
    # 1.4. two zero-mass particles
    space.clear_particles()
    space.add_particle(np.array((0., 0., 0.)), np.array((0., 0., 0.)), 0.)
    space.add_particle(np.array((1., 0., 0.)), np.array((0., 0., 0.)), 0.)
    accs = bf.calculate_accs_pp(space.r, space.m, G, eps)
    np.testing.assert_equal(accs, np.array([[0., 0., 0.], [0., 0., 0.]]))

    # 2. two particles on x-axis at distance 1.0
    space.clear_particles()
    space.add_particle(np.array((0., 0., 0.)), np.array((0., 0., 0.)), 1.0)
    space.add_particle(np.array((1., 0., 0.)), np.array((0., 0., 0.)), 1.0)
    accs = bf.calculate_accs_pp(space.r, space.m, G, eps)
    np.testing.assert_almost_equal(accs, np.array([[1., 0., 0.], [-1., 0., 0.]]))
    # 3. two particles of unequal masses on x-axis at distance 1.0
    space.clear_particles()
    space.add_particle(np.array((0., 0., 0.)), np.array((0., 0., 0.)), 1.0)
    space.add_particle(np.array((1., 0., 0.)), np.array((0., 0., 0.)), 0.5)
    accs = bf.calculate_accs_pp(space.r, space.m, G, eps)
    np.testing.assert_almost_equal(accs, np.array([[0.5, 0., 0.], [-1, 0., 0.]]))



def test_barnes_hut_theta_zero_C():
    # test Barnes-Hut C kernel relative to Cython brute force kernel, for theta = 0
    G = 1.0
    eps = 1.0e-3
    theta = 0.
    particle_nums = [2, 10, 100, 1000, 5000, 10000]    
    for num in particle_nums:
        cube_length = int(np.sqrt(num))
        space = Space()
        space.add_cuboid(num, np.array((0., 0., 0.)), cube_length, cube_length, cube_length, vel_func, mass_func)
        accs1 = bf.calculate_accs_pp(space.r, space.m, G, eps)
        accs2 = oct.calc_accs_octree(cube_length, 0., 0., 0., space.r, space.m, G, eps, theta)
        np.testing.assert_almost_equal(accs1, accs2)



def test_barnes_hut_theta_non_zero_C():
    # test Barnes-Hut C kernel relative to Cython brute force kernel, for theta > 0
    G = 1.0
    eps = 1.0e-3
    theta = 0.5
    particle_nums = [2, 10, 100, 1000, 5000, 10000]
    for num in particle_nums:
        cube_length = int(np.sqrt(num))
        space = Space()
        space.add_cuboid(num, np.array((0., 0., 0.)), cube_length, cube_length, cube_length, vel_func, mass_func)
        accs2 = bf.calculate_accs_pp(space.r, space.m, G, eps)
        accs3 = oct.calc_accs_octree(cube_length, 0., 0., 0., space.r, space.m, G, eps, theta)
        err, std_err = calculate_relative_error(accs3, accs2)
        assert err < 0.02 and std_err < 0.02



cimport cython
cimport numpy as np
import numpy as np
from libc.math cimport sqrt
from cython.parallel import prange
cimport openmp

ctypedef np.float64_t DTYPE_t


# ---------------------------------------------------
# Cython brute force
# ---------------------------------------------------

@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calculate_accs_pp(DTYPE_t [:, :] r, DTYPE_t[:] m, double G, double eps):
	cdef int n = r.shape[0]
	cdef int k = r.shape[1]
	cdef DTYPE_t [:, :] accs = np.zeros([n, k])
	cdef int i, j, p
	cdef double d, f, dx, dy, dz, ds, k1, k2
	cdef int n_threads = openmp.omp_get_max_threads()
	cdef int start_ind, stop_ind
	cdef int chunk_size = int(n / n_threads)
	cdef DTYPE_t [:, :, :] temp = np.zeros([n, 6, k])

	for p in prange(n_threads, nogil=True):
		start_ind = p * chunk_size
		stop_ind = (p + 1) * chunk_size if p < n_threads - 1 else n
		for i in range(start_ind, stop_ind):
			k1 = m[i] * G
			for j in range(i + 1, n):
				k2 = m[j] * G
				dx = r[j, 0] - r[i, 0]
				dy = r[j, 1] - r[i, 1]
				dz = r[j, 2] - r[i, 2]
				ds = dx * dx + dy * dy + dz * dz
				d = sqrt(ds)
				f = 1.0 / (d * d * d + eps)
				# update accs
				accs[i, 0] += dx * f * k2
				accs[i, 1] += dy * f * k2
				accs[i, 2] += dz * f * k2
				# update temporary array - we need this to avoid thread race
				temp[j, p, 0] -= dx * f * k1
				temp[j, p, 1] -= dy * f * k1
				temp[j, p, 2] -= dz * f * k1

	accs += np.sum(temp, axis=1)

	return np.asarray(accs)


# ---------------------------------------------------
# Wrappers
# ---------------------------------------------------

def calculate_accs_pp(r, m, G, eps):
	"""
	Parallel brute-force calculation of accelerations.

	:param r: N x 3 numpy array, position vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:param G: gravitational constant
	:param eps: gravitational softening
	:return: N x 3 numpy array, acceleration vectors of particles
	"""
	return _calculate_accs_pp(r, m, G, eps)



import time
import numpy as np
from functools import wraps


# --------------------------------------
# General utility functions
# --------------------------------------

def timing(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = f(*args, **kwargs)
        end = time.time()
        t = end - start
        print('Function name: {0}, Elapsed time: {1}'.format(f.__name__, t))
        return result
    return wrapper


# --------------------------------------
# Numerical utility functions
# --------------------------------------

def calculate_relative_error(accs, accs_ref):
    error_vec = np.linalg.norm(accs - accs_ref, axis=1) / np.linalg.norm(accs_ref, axis=1)
    return np.mean(error_vec), np.std(error_vec)


def to_cartesian(r, coord_sys):
    """
    Transforms an array of vectors from a coord_sys to the cartesian coordinate
    system.
    :param r: n x 3 matrix of position or any other vectors in spherical
    coordinates
    :param coord_sys: 'spherical' or 'cylindrical'
    :return: n x 3 matrix of position vectors in cartesian coordinates
    """
    if coord_sys == 'cylindrical':
        r, phi, z = r[:, 0], r[:, 1], r[:, 2]
        x = r * np.cos(phi)
        y = r * np.sin(phi)
        return np.column_stack((x, y, z))
    elif coord_sys == 'spherical':
        r, theta, phi = r[:, 0], r[:, 1], r[:, 2]
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)
        return np.column_stack((x, y, z))
    else:
        raise ValueError("'coord_sys' must be 'cylindrical' or 'spherical'.")


# --------------------------------------
# Progress bar
# --------------------------------------

class ProgressBarStyle:
    def __init__(self, fill=u'\u2588', empty='.', left='[', right=']', text='Progress: '):
        """

        :param fill: symbol for completed percentage
        :param empty: symbol for remaining percentage
        :param left: left bound symbol
        :param right: right bound symbol
        :param text: text which is displayed before the progress bar
        """
        self.fill = fill
        self.empty = empty
        self.left = left
        self.right = right
        self.text = text


class ProgressBar:
    def __init__(self, iterations, length, style=ProgressBarStyle()):
        """

        :param iterations: total number of iterations
        :param length: length of progress bar
        :param style: instance of ProgressBarStyle, determines visual properties
        """
        self._iterations = iterations
        self._length = length
        self._style = style
        self._current_iter = 0

    def set_style(self, style):
        """
        Sets the style.

        :param style: instance of ProgressBarStyle
        """
        self._style = style

    def reset(self, iterations=None):
        """
        Resets the progress bar - sets the current iteration to 0, and, optionally, changes the total
        number of iterations.

        :param iterations: if given, resets the total number of iterations
        """
        self._current_iter = 0
        if iterations:
            self._iterations = iterations

    def update(self):
        """
        Updates the progress bar. An update consists of incrementing the current iteration, and
        displaying the progress bar. Also, it is checked if the current iteration is within specified bounds.
        """
        self._current_iter += 1
        assert self._current_iter <= self._iterations
        self._show()

    def _show(self):
        progress = self._current_iter / self._iterations
        filled_length = int(progress * self._length)
        empty_length = self._length - filled_length
        progress_percentage = str(int(progress * 100)) + ' %'
        print('\r{}{}{}{}{} {}'.format(self._style.text, self._style.left, self._style.fill * filled_length,
                                       self._style.empty * empty_length, self._style.right,
                                       progress_percentage), end='')
        if self._current_iter == self._iterations:
            print()



from datetime import datetime

from PyQt5.QtWidgets import (QMainWindow, QWidget, QPushButton, QLabel, QGroupBox,
                             QAction, QFileDialog, QApplication, QVBoxLayout, QSlider, QTabWidget,
                             QTableWidgetItem, QComboBox, QHBoxLayout, QTableView, QTableWidget)
from PyQt5.QtCore import Qt, QAbstractTableModel, QTimer, QModelIndex

import sys
from pathlib import Path

from gonzales.simulator.simulation import ResultReader

import vispy
import vispy.scene
from vispy.scene import visuals


class TableModel(QAbstractTableModel):
    _header_label_map = {'position': ['r_x', 'r_y', 'r_z'],
                         'velocity': ['v_x', 'v_y', 'v_z'],
                         'angular_momentum': ['L_x', 'L_y', 'L_z'],
                         'energy': ['E']}

    def __init__(self, data, res_name):
        super(TableModel, self).__init__()
        self._data = data
        self._res_name = res_name

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if len(self._data.shape) > 1:
                return str(self._data[index.row()][index.column()])
            return str(self._data[index.column()])

    def rowCount(self, parent=QModelIndex()):
        if len(self._data.shape) > 1:
            return self._data.shape[0]
        return 1

    def columnCount(self, parent=QModelIndex()):
        if len(self._data.shape) > 1:
            return self._data.shape[1]
        return self._data.shape[0]

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        labels = self._header_label_map.get(self._res_name)
        if labels and role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return labels[section]            
        return QAbstractTableModel.headerData(self, section, orientation, role)


class SliderLabelWidget(QWidget):
    def __init__(self, parent):
        super(SliderLabelWidget, self).__init__(parent)
        hbox = QHBoxLayout()

        self.sld = QSlider(Qt.Horizontal, self)
        self.sld.setRange(0, 100)
        self.sld.setFocusPolicy(Qt.NoFocus)
        self.sld.setPageStep(5)

        self.label = QLabel('0', self)

        hbox.addWidget(self.sld)
        hbox.addSpacing(15)
        hbox.addWidget(self.label)

        self.setLayout(hbox)
 

class WidgetWithLabel(QWidget):
    def __init__(self, parent, widget_inst, label_text):
        super(WidgetWithLabel, self).__init__(parent)
        hbox = QHBoxLayout()

        self.wgt = widget_inst
        self.label = QLabel(label_text, self)

        hbox.addWidget(self.label)
        hbox.addWidget(self.wgt)
        hbox.setAlignment(Qt.AlignLeft)
        
        self.setLayout(hbox)


class MainWidget(QWidget):
    def __init__(self, parent):
        super(MainWidget, self).__init__(parent)
        layout = QVBoxLayout()

        self.tabs = QTabWidget()
        self.tab1 = QWidget()
        self.tab2 = QWidget()
        self.tab3 = QWidget()

        self.tabs.addTab(self.tab1, "Animation")
        self.tabs.addTab(self.tab2, "Results")
        self.tabs.addTab(self.tab3, "Info")

        tab1_layout = QVBoxLayout()
        tab2_layout = QVBoxLayout()
        tab3_layout = QVBoxLayout()

        canvas = vispy.scene.SceneCanvas(keys='interactive', show=True)
        view = canvas.central_widget.add_view()
        view.camera = 'turntable'
        #view.camera.scale_factor = 1e11
        axis = vispy.scene.visuals.XYZAxis(parent=view.scene)
        self.scatter = visuals.Markers()
        view.add(self.scatter)
        self.view_widget = canvas.native

        play_button = QPushButton('Play', self)
        play_button.clicked.connect(self.parent().on_play)

        self.pause_button = QPushButton('Pause', self)
        self.pause_button.clicked.connect(self.parent().on_pause)
        
        # tab1
        tab1_layout.addWidget(play_button)
        tab1_layout.addWidget(self.pause_button)
        tab1_layout.addWidget(self.view_widget)
        self.step_label = QLabel()
        self.filename_label = QLabel()

        info_groupbox = QGroupBox("Simulation info")
        info_box = QVBoxLayout()
        info_box.addWidget(self.filename_label)
        info_box.addWidget(self.step_label)
        info_groupbox.setLayout(info_box)
        tab1_layout.addWidget(info_groupbox)
        self.tab1.setLayout(tab1_layout)

        # tab2
        self.combo = WidgetWithLabel(self, QComboBox(self), 'Result: ')


        self.table = QTableView()
        self.slider = WidgetWithLabel(self, SliderLabelWidget(self), 'Step: ')
        self.combo.wgt.activated.connect(self.parent().on_combo_activated)
        self.slider.wgt.sld.valueChanged.connect(self.parent().on_slider_changed)

        tab2_layout.addWidget(self.combo)
        tab2_layout.addWidget(self.slider)
        tab2_layout.addWidget(self.table)
        self.tab2.setLayout(tab2_layout)

        #tab3
        self.info_table = QTableWidget()
        self.info_table.verticalHeader().setVisible(False)
        self.info_table.horizontalHeader().setVisible(False)
        tab3_layout.addWidget(self.info_table)
        self.tab3.setLayout(tab3_layout)

        layout.addWidget(self.tabs)
        self.setLayout(layout)


class NBodyViewer(QMainWindow):
    def __init__(self, filename='', body_sizes=4, colors=(.9, .9, .1, .7)):
        super().__init__()
        self._filename = None
        self._reader = None
        self._timer = QTimer()
        self._cnt = 0
        self._num_steps = 0
        self._color = colors        
        self._sizes = body_sizes

        self.init_ui()
        if filename:
            self._set_data_from_file(filename)
        self._is_playing = False

    def _refresh_status_bar(self):
        self.main_widget.filename_label.setText('Loaded file: {}'.format(self._filename))

    def init_ui(self):
        self.main_widget = MainWidget(self)
        self.setCentralWidget(self.main_widget)
        self.statusbar = self.statusBar()
        self._refresh_status_bar()

        openFile = QAction('Open', self)
        openFile.setShortcut('Ctrl+O')
        openFile.setStatusTip('Open new File')

        play = QAction('Play', self)
        play.setStatusTip('Run animation')

        quit = QAction('Quit', self)
        quit.setStatusTip('Close viewer')

        openFile.triggered.connect(self.on_file_open)
        play.triggered.connect(self.on_play)
        quit.triggered.connect(self.close)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(openFile)
        fileMenu.addAction(play)
        fileMenu.addAction(quit)

        self.setGeometry(50, 50, 1600, 900)
        self.setWindowTitle('N-body visualization')
        self.show()

    def _update_view(self):
        self._cnt += 1
        if self._cnt > self._num_steps:
            return
        pos_data = self._reader.get_result('position', self._cnt)
        self.main_widget.scatter.set_data(pos_data, edge_color=None, face_color=self._color, size=self._sizes)
        self.main_widget.step_label.setText('Step: {}'.format(self._cnt))

    def _populate_info_table(self):
        n = len(self._reader.get_info().keys())
        self.main_widget.info_table.setRowCount(n)
        self.main_widget.info_table.setColumnCount(2)
        for i, kname in enumerate(self._reader.get_info().keys()):
            val = self._reader.get_info()[kname][()]
            self.main_widget.info_table.setItem(i, 0, QTableWidgetItem(str(kname)))
            if kname in ['start_time', 'end_time']:
                val = datetime.fromtimestamp(val).strftime("%m/%d/%Y, %H:%M:%S")
            self.main_widget.info_table.setItem(i, 1, QTableWidgetItem(str(val)))
        self.main_widget.info_table.resizeColumnsToContents()

    def on_play(self):
        if not self._filename:
            self.statusBar().showMessage('No file loaded')
            return
        self._cnt = 0
        self._timer.timeout.connect(self._update_view)
        self._timer.start(50)
        self._is_playing = True

    def on_pause(self):
        if self._is_playing:
            self._is_playing = False
            self.main_widget.pause_button.setText('Resume')
            self._timer.stop()
            return
        self._timer.start(50)
        self._is_playing = True
        self.main_widget.pause_button.setText('Pause')

    def on_file_open(self):
        home_dir = str(Path.home())
        fname = QFileDialog.getOpenFileName(self, 'Open file', home_dir)
        if not any(fname):
            return
        self._cnt = 0
        if self._reader:
            self._reader.close()
        self._sizes = 4
        self._set_data_from_file(fname[0])

    def on_combo_activated(self):
        res_name = str(self.main_widget.combo.wgt.currentText())
        res_data = self._reader.get_result(res_name, 0)
        self.main_widget.table.setModel(TableModel(res_data, res_name))
        self.main_widget.slider.wgt.sld.setRange(0, self._reader.get_result_num_steps(res_name) - 1)
        self.main_widget.slider.wgt.sld.setValue(0)

    def on_slider_changed(self, value):
        res_name = str(self.main_widget.combo.wgt.currentText())
        res_data = self._reader.get_result(res_name, value)
        res_freq = self._reader.get_result_frequency(res_name)
        self.main_widget.slider.wgt.label.setText(str(res_freq * value))
        self.main_widget.slider.wgt.sld.setValue(value)
        self.main_widget.table.setModel(TableModel(res_data, res_name))

    def _set_data_from_file(self, filename):
        self._filename = filename
        self._reader = ResultReader(filename)
        pos_data = self._reader.get_result('position', self._cnt)
        self.main_widget.scatter.set_data(pos_data, edge_color=None, face_color=self._color, size=self._sizes)
        self._refresh_status_bar()
        self._num_steps = self._reader.get_info()['number_of_steps'][()]
        self._populate_info_table()
        self.main_widget.combo.wgt.clear()
        for res_name in self._reader.get_result_names():
            self.main_widget.combo.wgt.addItem(res_name)
        res_name = str(self.main_widget.combo.wgt.currentText())
        res_data = self._reader.get_result(res_name, 0)
        self.main_widget.table.setModel(TableModel(res_data, res_name))
        self.main_widget.slider.wgt.sld.setRange(0, self._reader.get_result_num_steps(res_name) - 1)

    def closeEvent(self, event):
        self._timer.stop()
        if self._reader:
            self._reader.close()
        super(QMainWindow, self).closeEvent(event)


def run_viewer(filename='', body_sizes=4, colors=(.9, .9, .1, .7)):
    """
    Runs the viewer with given filename.

    :param filename: filepath of hdf5 file
    :param body_sizes: int or tuple of ints, respective body size(s) which will be drawn.
    :param colors: double or tuple of doubles, respective body color(s) which will be drawn.
    """
    app = QApplication(sys.argv)
    nbv = NBodyViewer(filename, body_sizes, colors)
    sys.exit(app.exec_())


if __name__ == '__main__':
    run_viewer()



cimport cython
cimport numpy as np
import numpy as np
from libc.math cimport sqrt
from cython.parallel import prange
cimport openmp


ctypedef np.float64_t DTYPE_t


# ---------------------------------------------------
# Physics functions
# ---------------------------------------------------

@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calc_com(DTYPE_t [:, :] r, DTYPE_t [:] m):
	"""
		Calculate center of mass.
	"""
	cdef int n = r.shape[0]
	cdef DTYPE_t [:] com = np.zeros(3)
	cdef int i
	cdef double mass = 0

	for i in range(n):
		mass += m[i]
		com[0] += m[i] * r[i, 0]
		com[1] += m[i] * r[i, 1]
		com[2] += m[i] * r[i, 2]
	com[0] /= mass
	com[1] /= mass
	com[2] /= mass
	return np.asarray(com)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calc_ke(DTYPE_t [:, :] v, DTYPE_t [:] m):
	"""
		Calculate kinetic energy.
	"""
	cdef int n = v.shape[0]
	cdef int i
	cdef double ke = 0

	for i in range(n):
		ke += m[i] * (v[i, 0] * v[i, 0] + v[i, 1] * v[i, 1] + v[i, 2] * v[i, 2])
	return 0.5 * ke


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calc_pe(DTYPE_t [:, :] r, DTYPE_t [:] m, double G, double eps):
	"""
		Calculate potential energy.
	"""
	cdef int n = r.shape[0]
	cdef int i, j
	cdef double dx, dy, dz, ds, d
	cdef double pe = 0

	for i in prange(n, nogil=True):
		for j in range(i + 1, n):
			dx = r[j, 0] - r[i, 0]
			dy = r[j, 1] - r[i, 1]
			dz = r[j, 2] - r[i, 2]
			ds = dx * dx + dy * dy + dz * dz
			d = sqrt(ds)
			pe += m[i] * m[j] / (d + eps)
	return - G * pe


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calc_te(DTYPE_t [:, :] r, DTYPE_t [:, :] v, DTYPE_t [:] m, double G, double eps):
	"""
		Calculates total energy.
	"""
	cdef double te
	te = _calc_pe(r, m, G, eps) + _calc_ke(v, m)
	return te


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _calc_ang_momentum(DTYPE_t [:, :] r, DTYPE_t [:, :] v, DTYPE_t [:] m):
	"""
		Calculates angular momentum.
	"""
	cdef int n = r.shape[0]
	cdef DTYPE_t [:] angm = np.zeros(3)
	cdef int i

	for i in range(n):
		angm[0] += m[i] * (r[i, 1] * v[i, 2] - r[i, 2] * v[i, 1])
		angm[1] += m[i] * (r[i, 2] * v[i, 0] - r[i, 0] * v[i, 2])
		angm[2] += m[i] * (r[i, 0] * v[i, 1] - r[i, 1] * v[i, 0])
	return angm


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _advance_r(DTYPE_t [:, :] r, DTYPE_t [:, :] v, DTYPE_t [:, :] accs, double dt):
	"""
		Updates positions for time step dt.
	"""
	cdef int n = accs.shape[0]
	cdef int i

	for i in range(n):
		r[i, 0] +=  v[i, 0] * dt + 0.5 * accs[i, 0] * dt * dt
		r[i, 1] +=  v[i, 1] * dt + 0.5 * accs[i, 1] * dt * dt
		r[i, 2] +=  v[i, 2] * dt + 0.5 * accs[i, 2] * dt * dt


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _advance_v(DTYPE_t [:, :] v, DTYPE_t [:, :] accs, DTYPE_t [:, :] new_accs, double dt):
	"""
		Updates velocities for time step dt.
	"""
	cdef int n = accs.shape[0]
	cdef int i

	for i in range(n):
		v[i, 0] +=  0.5 * (accs[i, 0] + new_accs[i, 0]) * dt
		v[i, 1] +=  0.5 * (accs[i, 1] + new_accs[i, 1]) * dt
		v[i, 2] +=  0.5 * (accs[i, 2] + new_accs[i, 2]) * dt


# ---------------------------------------------------
# Wrappers
# ---------------------------------------------------

def calc_com(r, m):
	"""
	Calculates center of mass of a set of particles.

	:param r: N x 3 numpy array, position vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:return: 1 x 3 numpy array, center of mass
	"""
	return _calc_com(r, m)


def calc_ke(v, m):
	"""
	Calculates kinetic energy of a system of particles.

	:param v: N x 3 numpy array, velocity vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:return: double, kinetic energy of the system
	"""
	return _calc_ke(v, m)


def calc_pe(r, m, G, eps):
	"""
	Calculates potential energy of a system of particles.

	:param r: N x 3 numpy array, position vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:param G: gravitational constant
	:param eps: gravitational softening
	:return: double, potential energy of the system
	"""
	return _calc_pe(r, m, G, eps)


def calc_te(r, v, m, G, eps):
	"""
	Calculates total mechanical energy of a system of particles.

	:param r: N x 3 numpy array, position vectors of the particles
	:param v: N x 3 numpy array, velocity vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:param G: gravitational constant
	:param eps: gravitational softening
	:return: double, total mechanical energy of the system
	"""
	return _calc_te(r, v, m, G, eps)


def calc_ang_mom(r, v, m):
	"""
	Calculates angular momentum of a system of particles.

	:param r: N x 3 numpy array, position vectors of the particles
	:param v: N x 3 numpy array, velocity vectors of the particles
	:param m: N x 1 numpy array, mass vector of the particles
	:return: 1 x 3 numpy array, angular momentum vector
	"""
	return _calc_ang_momentum(r, v, m)


def advance_r(r, v, accs, dt):
	_advance_r(r, v, accs, dt)


def advance_v(v, accs, new_accs, dt):
	_advance_v(v, accs, new_accs, dt)



import os
import sys
import numpy as np
from pathlib import Path
from Cython.Build import cythonize
from setuptools import setup, Extension, find_packages


CYTHON_SRC_PATH = os.path.join('src', 'cython')
C_SRC_PATH = os.path.join('src', 'c')
PYTHON_SRC_PATH = os.path.join('src', 'python')

long_description = (Path(__file__).parent / "README.md").read_text()


def get_compile_args():
    if sys.platform == 'linux':
        return ['-fopenmp', '-O3', '-ffast-math', '-march=native']
    elif sys.platform == 'win32':
        return ['/openmp']
    return []


def get_link_args():
    if sys.platform == 'linux':
        return ['-fopenmp']
    elif sys.platform == 'win32':
        return []
    return []
    

compile_args = get_compile_args()
link_args = get_link_args()


extensions = [Extension("gonzales.lib.brute_force", sources=[os.path.join(CYTHON_SRC_PATH, 'brute_force.pyx')],
                        extra_compile_args=compile_args,
                        extra_link_args=link_args,
                        include_dirs=[CYTHON_SRC_PATH, np.get_include()]),
              Extension("gonzales.lib.octree", sources=[os.path.join(CYTHON_SRC_PATH, 'octree.pyx'),
                                                        os.path.join(C_SRC_PATH, 'data_structs.c'),
                                                        os.path.join(C_SRC_PATH, 'octnode.c'),
                                                        os.path.join(C_SRC_PATH, 'brute_force.c')],
                        include_dirs=[CYTHON_SRC_PATH, C_SRC_PATH, np.get_include()],
                        extra_compile_args=compile_args,
                        extra_link_args=link_args,
                        language="c"),

              Extension("gonzales.lib.physics", sources=[os.path.join(CYTHON_SRC_PATH, 'physics.pyx')],
                        extra_compile_args=compile_args,
                        extra_link_args=link_args,
                        include_dirs=[C_SRC_PATH, np.get_include()]),
              ]


setup(
    name='gonzales',
    version='0.1.0',
    packages=find_packages(PYTHON_SRC_PATH),
    package_dir={'': PYTHON_SRC_PATH},
    description='N-body simulator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Radovan Horvat',
    author_email='radovan.horvat@gmail.com',
    url='https://github.com/radovanhorvat/gonzales',
    license='MIT',
    install_requires=['numpy>=1.21.0', 'Cython>=0.29.23', 'h5py>=3.3.0', 'psutil>=5.8.0', 'vispy>=0.7.1',
                      'PyQt5>=5.15.4'],
    ext_modules=cythonize(extensions),
)



from gonzales.simulator.space import Space
from gonzales.simulator.simulation import PPSimulation, BHSimulation, ResultReader
from gonzales.ui.viewer import run_viewer
from gonzales.benchmark.benchmark import run_default_benchmark



import os
import h5py
import shutil
import numpy as np
from copy import deepcopy
from contextlib import contextmanager

from gonzales.simulator.space import Space
import gonzales.simulator.simulation as sim


# ---------------------------------------------------------
# Utility functions
# ---------------------------------------------------------

@contextmanager
def make_temp_result_dir(dir_name):
    dir_path = os.path.dirname(os.path.join(os.path.abspath(__file__)))
    temp_dir = os.path.join(dir_path, dir_name)
    os.mkdir(temp_dir)
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


def _are_results_equal(sim1, sim2, n_steps):
    with h5py.File(sim1.output_filepath, 'r') as pp, h5py.File(sim2.output_filepath, 'r') as bh:
        pos_data_pp = pp['results/position']
        pos_data_bh = bh['results/position']
        vel_data_pp = pp['results/velocity']
        vel_data_bh = bh['results/velocity']
        for k in range(n_steps + 1):
            np.testing.assert_equal(pos_data_bh[k],  pos_data_pp[k])
            np.testing.assert_equal(vel_data_bh[k], vel_data_pp[k])


def _are_results_close(sim1, sim2, n_steps):
    with h5py.File(sim1.output_filepath, 'r') as pp, h5py.File(sim2.output_filepath, 'r') as bh:
        pos_data_pp = pp['results/position']
        pos_data_bh = bh['results/position']
        vel_data_pp = pp['results/velocity']
        vel_data_bh = bh['results/velocity']
        for k in range(n_steps + 1):
            assert _are_vecs_close(pos_data_bh[k], pos_data_pp[k])
            assert _are_vecs_close(vel_data_bh[k], vel_data_pp[k])


def _are_vecs_close(vec, vec_ref, rtol=1e-2, atol=1e-15):
    """

    :param vec: estimated vec
    :param vec_ref: vec to which we compare (correct one)
    :param rtol: relative tolerance
    :param atol: absolute tolerance
    :return: bool
    """
    x = np.mean(np.linalg.norm(vec - vec_ref, axis=1))
    y = np.mean(np.linalg.norm(vec_ref, axis=1) * rtol + atol)
    return x <= y


# ---------------------------------------------------------
# Tests
# ---------------------------------------------------------

def vel_func(pos_vec):
    return np.array((0, 0, 0))


def mass_func(pos_vec):
    return 1.0


def test_sim_barnes_hut_theta_zero():
    # test Barnes-Hut simulation relative to Cython brute force simulation, for theta = 0
    with make_temp_result_dir('temp_results') as res_dir:
        pp_file = os.path.join(res_dir, 'test_pp.hdf5')
        bh_file = os.path.join(res_dir, 'test_bh.hdf5')
        G = 1.0
        eps = 1.0e-3
        theta = 0.
        n_steps = 100
        step_size = 0.01
        n = 1000
        cube_length = int(np.sqrt(n))
        space = Space()
        space.add_cuboid(n, np.array((0., 0., 0.)), cube_length, cube_length, cube_length, vel_func, mass_func)
        space1 = deepcopy(space)
        sim_pp = sim.PPSimulation(space, pp_file, G, eps)
        sim_bh = sim.BHSimulation(space1, bh_file, G, eps, cube_length, np.array((0., 0., 0.)), theta)
        sim_pp.run(n_steps, step_size)
        sim_bh.run(n_steps, step_size)
        # result comparsion
        _are_results_equal(sim_pp, sim_bh, n_steps)


def test_sim_barnes_hut_theta_non_zero():
    # test Barnes-Hut simulation relative to Cython brute force simulation, for theta = 0.5
    with make_temp_result_dir('temp_results') as res_dir:
        pp_file = os.path.join(res_dir, 'test_pp.hdf5')
        bh_file = os.path.join(res_dir, 'test_bh.hdf5')
        G = 1.0
        eps = 1.0e-3
        theta = 0.5
        n_steps = 100
        step_size = 0.01
        n = 1000
        cube_length = int(np.sqrt(n))
        space = Space()
        space.add_cuboid(n, np.array((0., 0., 0.)), cube_length, cube_length, cube_length, vel_func, mass_func)
        space1 = deepcopy(space)
        sim_pp = sim.PPSimulation(space, pp_file, G, eps)
        sim_bh = sim.BHSimulation(space1, bh_file, G, eps, 10 * cube_length, np.array((0., 0., 0.)), theta)
        sim_pp.run(n_steps, step_size)
        sim_bh.run(n_steps, step_size)
        # result comparsion
        _are_results_close(sim_pp, sim_bh, n_steps)


def test_energy_conservation_barnes_hut():
    # test Barnes-Hut energy conservation
    with make_temp_result_dir('temp_results') as res_dir:
        bh_file = os.path.join(res_dir, 'test_bh.hdf5')
        n = 1000
        G = 1.0
        eps = 1.0e-3
        theta = 0.75
        n_steps = 1000
        step_size = 0.001
        cube_length = int(np.sqrt(n))
        space = Space()
        space.add_cuboid(n, np.array((0., 0., 0.)), cube_length, cube_length, cube_length, vel_func, mass_func)
        sim_bh = sim.BHSimulation(space, bh_file, G, eps, 10 * cube_length, np.array((0., 0., 0.)), theta)
        sim_bh.add_result('energy', n_steps)
        sim_bh.run(n_steps, step_size)
        with h5py.File(sim_bh.output_filepath, 'r') as bh:
            e_in = bh['results']['energy'][0]
            e_fin = bh['results']['energy'][1]
            assert np.abs(e_in - e_fin) / np.abs(e_in) <= 0.01
